#!/usr/bin/env python
#
# StuDNA tools, version 2009-Mar
# (C) Martin Saturka www.bioplexity.org
# Available under the MIT/X license:
#
# Produces sqlite database of gene sequence
# info of an organism of a given taxon id
#

# gene2refseq.gz, taken positions:
#
# 0 tax_id 
# 1 gene_id
# 2 state
# 7 nc_acc.version
# 9 start_pos .. zero based!
#10 end_pos .. zero based!
#11 orientation
#12 assembly

# gene_info.gz
#
# 0 tax_id
# 1 gene_id
# 2 inf. symbol
# 8 desc
# 9 type
#10 auth. symbol

# chromosome_names.txt
#
# 0 chr. symbol
# 1 chr. type
# 2 chr. name
# 3 file_name

import sys, os, time
import getopt
import csv as C

withsqlite3 = 1
try:
  import sqlite3 as S
except:
  withsqlite3 = 0
if 0 == withsqlite3:
  try:
    from pysqlite2 import dbapi2 as S
  except:
    print("sqlite3 support has to be available")
    sys.exit(1)

class StudnaRed(object):
  def __init__(self):
    self.grefname = ""
    self.ginfname = ""
    self.chrfname = ""

    self.sqlitename = ""
    self.taxonid = "0"
    self.desctxt = ""

    self.totalcount = 0
    self.withoutname = 0

    self.error = 0

  def work(self):
    if not 0 == self.error:
      return False
    if not self.taxonid.isdigit():
      self.error = 1
      return False
    if "0" == self.taxonid:
      self.error = 2
      return False
    if "" == self.grefname:
      self.error = 3
      return False
    if "" == self.ginfname:
      self.error = 4
      return False
    if "" == self.chrfname:
      self.error = 5
      return False
    if "" == self.sqlitename:
      self.error = 6
      return False

    chromlines = []
    chromlist = []
    chromnamesfh = []
    [chrsymcol, chrtypcol, chrnamcol, chrfilcol] = [0, 1, 2, 3]
    try:
      chromnamesfh = open(self.chrfname)
    except:
      self.error = 101
      return False
    try:
      for line in chromnamesfh.readlines():
        line = line.strip()
        if 0 == len(line):
          continue
        if line.startswith("#"):
          continue
        onelist = line.split()
        if not 4 == len(onelist):
          continue
        curchrname = onelist[chrnamcol]
        chromlines.append(onelist)
        chromlist.append(curchrname)
      chromnamesfh.close()
    except:
      chromnamesfh.close()
      self.error = 102
      return False

    gnames = {}
    self.totalcount = 0
    self.withoutname = 0

    grefreqlen = 13
    ginfreqlen = 15
    [taxcol, gencol, symcol, descol, typcol, namcol] = [0, 1, 2, 8, 9, 10]
    [concol, chrcol, begcol, endcol, oricol] = [2, 7, 9, 10, 11]

    dbcon = []
    dbcur = []
    try:
      dbcon = S.connect(self.sqlitename)
      dbcur = dbcon.cursor()
    except:
      self.error = 201
      return False

    genetable = "genes" + self.taxonid
    sequtable = "sequences" + self.taxonid

    try:
      qcreate = "CREATE TABLE IF NOT EXISTS globinfo (id INTEGER PRIMARY KEY, "
      qcreate += "taxon TEXT, type INTEGER, desc TEXT, genestate INTEGER, sequstate INTEGER, "
      qcreate += "genetable TEXT UNIQUE, sequtable TEXT UNIQUE, "
      qcreate += "uplength INTEGER, downlength INTEGER, "
      qcreate += "CONSTRAINT taxtyp UNIQUE(taxon, type))"
      dbcur.execute(qcreate)
      dbcon.commit()
      qcreate = "CREATE TABLE IF NOT EXISTS chromosomes (id INTEGER PRIMARY KEY, "
      qcreate += "taxon TEXT, symbol INTEGER, type TEXT, name TEXT UNIQUE, file TEXT, "
      qcreate += "CONSTRAINT taxsym UNIQUE(taxon, symbol), "
      qcreate += "CONSTRAINT typval CHECK (type==\"lin\" OR type==\"cir\"))"
      dbcur.execute(qcreate)
      dbcon.commit()
      qcreate = "CREATE TABLE " + genetable + " "
      qcreate += "(gene INTEGER PRIMARY KEY, name TEXT UNIQUE, symbol TEXT, state TEXT, "
      qcreate += "chrom TEXT, start INTEGER, end INTEGER, orient TEXT, desc TEXT, type TEXT, "
      qcreate += "CONSTRAINT orival CHECK (orient==\"+\" OR orient==\"-\"))"
      dbcur.execute(qcreate)
      dbcon.commit()
      qcreate = "CREATE TABLE " + sequtable + " "
      qcreate += "(gene INTEGER PRIMARY KEY, upseq TEXT, downseq TEXT)"
      dbcur.execute(qcreate)
      dbcon.commit()
    except:
      dbcur.close()
      dbcon.close()
      self.error = 202
      return False

    try:
      qinsert = "INSERT INTO globinfo (id, taxon, type, desc, genestate, sequstate, "
      qinsert += "genetable, sequtable, uplength, downlength) "
      qinsert += "VALUES (NULL, ?, 1, ?, 0, 0, ?, ?, 0, 0)"
      params = (self.taxonid, self.desctxt, genetable, sequtable)
      dbcur.execute(qinsert, params)
      dbcon.commit()
    except:
      dbcur.close()
      dbcon.close()
      self.error = 203
      return False

    qinsert = "INSERT INTO chromosomes (id, taxon, symbol, type, name, file) "
    qinsert += "VALUES (NULL, ?, ?, ?, ?, ?)"
    for line in chromlines:
      cursymbol = line[chrsymcol]
      curtype = line[chrtypcol].lower()
      curname = line[chrnamcol]
      curfile = line[chrfilcol]
      params = (self.taxonid, cursymbol, curtype, curname, curfile)
      try:
        dbcur.execute(qinsert, params)
        dbcon.commit()
      except:
        dbcur.close()
        dbcon.close()
        self.error = 204
        return False

    try:
      fhginf = open(self.ginfname, "r")
    except:
      dbcur.close()
      dbcon.close()
      self.error = 301
      return False

    crginf = C.reader(fhginf, delimiter="\t")
    for row in crginf:
      if len(row) < ginfreqlen:
        continue
      if row[0].startswith("#"):
        continue
      if not row[taxcol] == self.taxonid:
        continue
      curgene = row[gencol].strip()
      curname = row[namcol].strip()
      cursymb = row[symcol].strip()
      curdesc = row[descol].strip()
      curtype = row[typcol].strip()
      if 0 == len(curgene):
        continue
      if not curgene.isdigit():
        continue
      if 0 == len(curname):
        curname = "-"
      gnames[int(curgene)] = [curname, cursymb, curdesc, curtype]

    fhginf.close()
    [onenamecol, onesymbcol, onedesccol, onetypecol] = [0, 1, 2, 3]

    try:
      fhgref = open(self.grefname, "r")
    except:
      dbcur.close()
      dbcon.close()
      self.error = 302
      return False

    qinsert = "INSERT INTO " + genetable + " "
    qinsert += "(gene, name, symbol, state, chrom, start, end, orient, desc, type) "
    qinsert += "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"

    gpassed = 0
    crgref = C.reader(fhgref, delimiter="\t")
    for row in crgref:
      if len(row) < grefreqlen:
        continue
      if row[0].startswith("#"):
        continue
      if not row[taxcol] == self.taxonid:
        continue
      curgene = row[gencol].strip()
      curcond = row[concol].strip()
      curchrm = row[chrcol].strip()
      curbegp = row[begcol].strip()
      curendp = row[endcol].strip()
      curorin = row[oricol].strip()

      isok = True
      for item in [curgene, curcond, curchrm, curbegp, curendp, curorin]:
        if 0 == len(item):
          isok = False
          break
      if not isok:
        continue

      isok = True
      for item in [curgene, curbegp, curendp]:
        if not item.isdigit():
          isok = False
          break
      if not isok:
        continue
      curgene = int(curgene)
      curbegp = int(curbegp)
      curendp = int(curendp)

      curcond = curcond.lower()

      if not curchrm in chromlist:
        continue
      if not curorin in ["+", "-"]:
        continue

      if curgene <= gpassed:
        continue
      gpassed = curgene
      self.totalcount += 1

      curname = "-"
      cursymb = ""
      curdesc = ""
      curtype = ""
      if curgene in gnames:
        onelist = gnames[curgene]
        curname = onelist[onenamecol]
        cursymb = onelist[onesymbcol]
        curdesc = onelist[onedesccol]
        curtype = onelist[onetypecol]
      if "-" == curname:
        curname = str(curgene)
        self.withoutname += 1

      params = (curgene, curname.upper(), cursymb.upper(), curcond, curchrm, curbegp, curendp, curorin, curdesc, curtype)
      try:
        dbcur.execute(qinsert, params)
      except:
        fhgref.close()
        dbcur.close()
        dbcon.close()
        self.error = 401
        return False

    fhgref.close()

    try:
      dbcon.commit()
    except:
      dbcur.close()
      dbcon.close()
      self.error = 501
      return False

    try:
      qupdate = "UPDATE globinfo SET genestate=1 WHERE taxon=? AND type=1"
      params = (self.taxonid,)
      dbcur.execute(qupdate, params)
      dbcon.commit()
    except:
      dbcur.close()
      dbcon.close()
      self.error = 601
      return False

    dbcur.close()
    dbcon.close()

    return True

#params: -r refname, -i infname, -c chromfile, -s sqlitename, -t taxonid, -d description
if ('__main__' == __name__):
  args = sys.argv[1:]
  argl = len(args)
  runstring = "Should be run: studna-red.py -r gene2refseq_file -i gene_inf_file -c chromosome_names_file "
  runstring += "-s output_database_file -t taxon_id (e.g. 9606 for h. sapiens) -d description\n"

  refname = ""
  infname = ""
  chrname = ""
  sqlname = ""
  taxonid = ""
  desctxt = ""

  optlist = []
  try:
    optlist, args = getopt.getopt(args, 'r:i:c:s:t:d:')
    for opt, arg in optlist:
      if "-r" == opt:
        refname = arg
      if "-i" == opt:
        infname = arg
      if "-c" == opt:
        chrname = arg
      if "-s" == opt:
        sqlname = arg
      if "-t" == opt:
        taxonid = arg
      if "-d" == opt:
        desctxt = arg

  except:
    sys.stderr.write(runstring)
    sys.exit(1)

  isok = True
  for item in [refname, infname, chrname, sqlname, taxonid]:
    if 0 == len(item):
      isok = False

  if not taxonid.isdigit():
    isok = False

  if not isok:
    sys.stderr.write(runstring)
    sys.exit(1)

  theRed = StudnaRed()
  theRed.grefname = refname
  theRed.ginfname = infname
  theRed.chrfname = chrname
  theRed.sqlitename = sqlname
  theRed.taxonid = taxonid
  theRed.desctxt = desctxt

  rv = theRed.work()
  if not rv:
    resmsg = "work failed, error: " + str(theRed.error)
    print(resmsg)
    sys.exit(1)

  tcount = str(theRed.totalcount)
  wcount = str(theRed.withoutname)
  resmsg = "work done, " + tcount + " genes, " + wcount + " without auth. name"
  print(resmsg)

