#!/usr/bin/env python
#
# StuDNA tools, version 2009-Mar
# (C) Martin Saturka www.bioplexity.org
# Available under the MIT/X license:
#
# Adds upstream sequences to the database
# prepared by the 'studna-red' script.
#

import sys, os, time
import getopt

withsqlite3 = 1
try:
  import sqlite3 as S
except:
  withsqlite3 = 0
if 0 == withsqlite3:
  try:
    from pysqlite2 import dbapi2 as S
  except:
    print("sqlite3 support has to be available")
    sys.exit(1)

class StudnaGreen(object):
  def __init__(self):
    self.sqlitename = ""
    self.taxonid = "0"

    self.chromdir = ""

    self.linelength = 70
    self.uplength = 2000
    self.downlength = 100

    self.seqcount = 0
    self.error = 0

  def work(self):
    if not 0 == self.error:
      return False
    if not self.taxonid.isdigit():
      self.error = 1
      return False
    if "0" == self.taxonid:
      self.error = 2
      return False
    if 0 == len(self.chromdir):
      self.error = 3
      return False
    if "" == self.sqlitename:
      self.error = 4
      return False
    for item in [self.linelength, self.uplength, self.downlength]:
      try:
        item += 0
      except:
        self.error = 5
        return False
      if 0 >= item:
        self.error = 6
        return False

    self.seqcount = 0

    linelengthskip = self.linelength + 1
    globtable = "globinfo"
    chrmtable = "chromosomes"
    genetable = ""
    sequtable = ""

    dbcon = []
    dbcur = []
    try:
      dbcon = S.connect(self.sqlitename)
      dbcur = dbcon.cursor()
    except:
      self.error = 101
      return False

    isok = True
    taxoninforows = 0
    [gestcol, sestcol, getacol, setacol] = [0, 1, 2, 3]
    try:
      qselect = "SELECT genestate, sequstate, genetable, sequtable FROM " + globtable + " WHERE taxon=? AND type=1"
      params = (self.taxonid,)
      dbcur.execute(qselect, params)
      for onerow in dbcur:
        taxoninforows += 1
        if not 1 == onerow[gestcol]:
          self.error = 201
          isok = False
        if not 0 == onerow[sestcol]:
          self.error = 202
          isok = False
        genetable = onerow[getacol]
        sequtable = onerow[setacol]
    except:
      self.error = 203
      dbcur.close()
      dbcon.close()
      return False

    if not 1 == taxoninforows:
      self.error = 301
      isok = False
    if 0 == len(genetable):
      self.error = 302
      isok = False
    if 0 == len(sequtable):
      self.error = 303
      isok = False

    if not isok:
      dbcur.close()
      dbcon.close()
      return False

    chromlist = []
    chromrows = []
    chromnamecol = 0
    try:
      qselect = "SELECT name, type, file FROM " + chrmtable + " WHERE taxon=? ORDER BY name ASC"
      params = (self.taxonid,)
      dbcur.execute(qselect, params)
      for onerow in dbcur:
        curchrom = onerow[chromnamecol]
        chromlist.append(curchrom)
        chromrows.append(onerow)
    except:
      self.error = 401
      dbcur.close()
      dbcon.close()
      return False

    generows = []
    try:
      qselect = "SELECT gene, chrom, start, end, orient FROM " + genetable + " ORDER BY chrom, start ASC"
      dbcur.execute(qselect)
      for onerow in dbcur:
        generows.append(onerow)
    except:
      self.error = 402
      dbcur.close()
      dbcon.close()
      return False

    try:
      dbcur.close()
      dbcon.close()
    except:
      self.error = 403
      return False

    revtable = {"0": "0", "A": "T", "T": "A", "G": "C", "C": "G"}
    totlen = self.uplength + self.downlength

    chromfiles = {}
    chromconts = {}
    [namcol, typcol, filcol] = [0, 1, 2]
    isok = True
    for onechr in chromrows:
      chrname = onechr[namcol]
      chrfile = onechr[filcol]
      seqfh = []
      firstpos = 0
      try:
        seqfh = open(os.path.join(self.chromdir, chrfile))
        seqfh.readline()
        firstpos = seqfh.tell()
      except:
        self.error = 501
        isok = False
      chromfiles[chrname] = seqfh
      chromconts[chrname] = firstpos

    seqfillbot = {}
    seqfilltop = {}
    for onechr in chromrows:
      if not isok:
        break

      chrname = onechr[namcol]
      chrtype = onechr[typcol]
      curpostop = ""
      curposbot = ""
      if "cir" == chrtype:
        circseq = ""
        seqh = chromfiles[chrname]
        for line in seqh.readlines():
          line = line.strip()
          if 0 == len(line):
            continue
          if line.startswith("#"):
            continue
          circseq += line
        circlen = len(circseq)
        if circlen >= totlen:
          curpostop = circseq[:totlen]
          curposbot = circseq[(circlen - totlen):]
        else:
          curpostop = circseq
          curpostop += (totlen - circlen) * "0"
          curposbot = (totlen - circlen) * "0"
          curposbot += circseq
        circseq = ""
      else:
        curpostop = totlen * "0"
        curposbot = totlen * "0"
      seqfillbot[chrname] = curposbot
      seqfilltop[chrname] = curpostop

    [gencol, chrcol, begcol, endcol, oricol] = [0, 1, 2, 3, 4]

    seqlines = []
    for onerow in generows:
      if not isok:
        break

      curgene = onerow[gencol]
      curchrom = onerow[chrcol]
      curstart = onerow[begcol]
      curend = onerow[endcol]
      curorient = onerow[oricol]

      if not curchrom in chromlist:
        continue
      self.seqcount += 1

      if not curorient in ["+", "-"]:
        self.error = 502
        isok = False
        break

      sequence = ""
      upseq = ""
      downseq = ""
      qinsert = "INSERT INTO " + sequtable + " (gene, upseq, downseq) VALUES (?, ?, ?)"

      charskip = 0
      curuplength = self.uplength
      curdownlength = self.downlength

      #this database contains 0-based positions, note that
      #visual web sources usually provide 1-based positions
      #start-end are border positions inside gene sequences
      if "+" == curorient:
        charskip = curstart - self.uplength
        if 0 > charskip:
          charless = -1 * charskip
          seqbot = seqfillbot[curchrom]
          sequence = seqbot[(totlen-charless):]
          charskip = 0
          curuplength -= charless
      if "-" == curorient:
        charskip = (curend + 1) - self.downlength
        if 0 > charskip:
          charless = -1 * charskip
          seqbot = seqfillbot[curchrom]
          sequence = seqbot[(totlen-charless):]
          charskip = 0
          curdownlength -= charless

      lineskip = charskip // self.linelength
      restskip = charskip % self.linelength

      seqfh = chromfiles[curchrom]
      firstpos = chromconts[curchrom]
      tomove = firstpos + (lineskip * linelengthskip)
      try:
        seqfh.seek(tomove)
      except:
        self.error = 601
        isok = False
        break

      readchars = restskip + curuplength + curdownlength
      readlines = readchars // self.linelength
      if 0 < (readchars % self.linelength):
        readlines += 1
      try:
        for ind in range(readlines):
          sequence += seqfh.readline().strip()
      except:
        self.error = 602
        isok = False
        break

      sequence = sequence[restskip:]
      if len(sequence) < totlen:
        seqtop = seqfilltop[curchrom]
        sequence += seqtop[:totlen - len(sequence)]

      if "-" == curorient:
        revsequence = sequence
        sequence = ""
        seqend = totlen - 1
        for ind in range(totlen):
          revchar = revsequence[seqend - ind]
          curchar = "N"
          if revchar in revtable:
            curchar = revtable[revchar]
          sequence += curchar

      upseq = sequence[:self.uplength].lower()
      downseq = sequence[self.uplength:totlen].lower()

      params = [curgene, upseq, downseq]
      seqlines.append(params)

    for onechr in chromfiles:
      try:
        seqfh = chromfiles[onechr]
        seqfh.close()
      except:
        isok = False

    if not isok:
      return False

    try:
      dbcon = S.connect(self.sqlitename)
      dbcur = dbcon.cursor()
    except:
      self.error = 701
      return False

    paramord = sorted(range(len(seqlines)))
    paramord.sort(key=lambda x: int(seqlines[x][0]))
    for ind in paramord:
      params = seqlines[ind]
      try:
        dbcur.execute(qinsert, tuple(params))
      except:
        dbcur.close()
        dbcon.close()
        self.error = 702
        return False

    try:
      qupdate = "UPDATE " + globtable + " SET sequstate=1, uplength=?, downlength=? WHERE taxon=? AND type=1"
      params = (self.uplength, self.downlength, self.taxonid)
      dbcur.execute(qupdate, params)
    except:
      self.error = 801
      dbcur.close()
      dbcon.close()
      return False

    try:
      dbcon.commit()
      dbcur.close()
      dbcon.close()
    except:
      self.error = 802
      return False

    return True


#params: -c chromdir, -l linelength, -u uplength, -d downlength, -s sqlitename, -t taxonid
if '__main__' == __name__:
  args = sys.argv[1:]
  argl = len(args)
  runstring = "Should be run: studna-green.py -c chromosome_directory -l fasta_line_length "
  runstring += "-u upstream_sequence_length -d downstream_sequence length "
  runstring += "-s output_database_file -t taxon_id (e.g. 9606 for h. sapiens)\n"

  chromdir = ""
  linelength = "70"
  uplength = "2000"
  downlength = "40"
  sqlitename = ""
  taxonid = ""

  optlist = []
  try:
    optlist, args = getopt.getopt(args, 'c:l:u:d:s:t:')
    for opt, arg in optlist:
      if "-c" == opt:
        chromdir = arg
      if "-l" == opt:
        linelength = arg
      if "-u" == opt:
        uplength = arg
      if "-d" == opt:
        downlength = arg
      if "-s" == opt:
        sqlitename = arg
      if "-t" == opt:
        taxonid = arg

  except:
    sys.stderr.write(runstring)
    sys.exit(1)

  isok = True
  for item in [chromdir, linelength, uplength, downlength, sqlitename, taxonid]:
    if 0 == len(item):
      isok = False

  for item in [linelength, uplength, downlength, taxonid]:
    if not item.isdigit():
      isok = False

  if not isok:
    sys.stderr.write(runstring)
    sys.exit(1)

  theGreen = StudnaGreen()
  theGreen.chromdir = chromdir
  theGreen.linelength = int(linelength)
  theGreen.uplength = int(uplength)
  theGreen.downlength = int(downlength)
  theGreen.sqlitename = sqlitename
  theGreen.taxonid = taxonid

  rv = theGreen.work()
  if not rv:
    resmsg = "work failed, error: " + str(theGreen.error)
    print(resmsg)
    sys.exit(1)

  resmsg = "work done, " + str(theGreen.seqcount) + " sequences"
  print(resmsg)

