#!/usr/bin/env Rscript
#
# StuDNA tools, version 2009-Jul
# (C) Martin Saturka www.bioplexity.org
# Available under the MIT/X license
#
# Search for similarities at requested sequences from
# the sqlite database prepared by the 'studna-yellow'
#
# Script parameters:
# sequence.fa    .. fasta formated sequences to be searched
# tuple_length   .. length of the subsite pairs (3 or 4 now)
# sep_lengths    .. colon separated list of gaps, e.g. 0:1:2
# estimates.csv  .. count probability values table
# incidences.csv .. result tuplet counts matrix
#

# libdir variable can hold the 'libbsites' library path, set e.g.
# libdir <- "/usr/local/lib/studna/"
libdir <- ""
dyn.load(paste(libdir, "libbsites", .Platform$dynlib.ext, sep=""), local=FALSE)

args <- commandArgs(TRUE)
if (5 != length(args)) {
  stop("should be run: studna-blue.R sequences.fa tuple_length sep_length_list output.csv")
}

fname <- args[1]
tuple.strlen <- args[2]
tuple.length <- 0
if ("3" == tuple.strlen) {
  tuple.length <- 3
}
if ("4" == tuple.strlen) {
  tuple.length <- 4
}
if (0 == tuple.length) {
  stop("unsupported tuple length")
}

sep.lengths <- as.integer(strsplit(args[3], ":")[[1]])
sep.cases <- length(sep.lengths)

non.lengths <- length(c(1:sep.cases)[is.na(sep.lengths)])
if (0 < non.lengths) {
  stop("non-numeric separation length(s)")
}
if (0 > min(sep.lengths)) {
  stop("negative separation length(s)")
}
sep.min = min(sep.lengths)

df <- read.table(fname, comment.char="#", check.names=FALSE, header=FALSE, sep="\n")
dfm <- as.matrix(df[[1]])
dfm.len <- dim(dfm)[1]

dfm.name.ind <- ('>' == substr(dfm[], 1, 1))

dfn.names <- dfm[dfm.name.ind]
dfn.index <- c(c(1:dfm.len)[dfm.name.ind], (dfm.len + 1))

dfm.cases <- length(dfn.names)
ftake <- function(x) {
  ind1 <- dfn.index[x] + 1
  ind2 <- dfn.index[x+1] - 1
  paste(dfm[ind1:ind2], sep="", collapse="")
}
dfn <- as.matrix(apply(as.matrix(c(1:dfm.cases)), 1, ftake))
dimnames(dfn) <- list(dfn.names, c("sequence"))

site.cases <- 4096
if (4 == tuple.length) {
  site.cases <- 65536
}

tcounts <- matrix(0, nrow=site.cases, ncol=(sep.cases * 3))
totlen <- sep.cases * site.cases * 3

cvec.maxlen <- 0
nnames <- c("a", "t", "g", "c")
fcount <- function(x) {
  cvec <- strsplit(tolower(dfn[x]), "")[[1]]
  cvec.len <- length(cvec)
  if (cvec.maxlen < cvec.len) {
    cvec.maxlen <<- cvec.len
  }

  ivec <- match(cvec, nnames, nomatch=0) - 1

  res <- .C("tuplecounts",
    as.integer(tuple.length),
    tups = integer(totlen),
    as.integer(length(ivec)),
    as.integer(ivec),
    as.integer(sep.cases),
    as.integer(sep.lengths),
    integer(length(ivec)),
    retval = integer(1)
  )

  curres <- res$tups
  dim(curres) <- c(site.cases, (sep.cases * 3))

  tcounts <<- tcounts + curres

  return()
}
nt <- apply(as.matrix(c(1:dfm.cases)), 1, fcount)

rn3t <- function(x) {
  x <- x - 1

  ind0 <- x %% 4
  ival <- x %/% 4
  ind1 <- ival %% 4
  ival <- ival %/% 4
  ind2 <- ival %% 4
  ival <- ival %/% 4
  ind3 <- ival %% 4
  ival <- ival %/% 4
  ind4 <- ival %% 4
  ival <- ival %/% 4
  ind5 <- ival %% 4

  val0 <- nnames[ind0 + 1]
  val1 <- nnames[ind1 + 1]
  val2 <- nnames[ind2 + 1]
  val3 <- nnames[ind3 + 1]
  val4 <- nnames[ind4 + 1]
  val5 <- nnames[ind5 + 1]

  paste(val5, val4, val3, ":", val2, val1, val0, sep="")
}

rn4t <- function(x) {
  x <- x - 1

  ind0 <- x %% 4

  ival <- x %/% 4
  ind1 <- ival %% 4
  ival <- ival %/% 4
  ind2 <- ival %% 4
  ival <- ival %/% 4
  ind3 <- ival %% 4
  ival <- ival %/% 4
  ind4 <- ival %% 4
  ival <- ival %/% 4
  ind5 <- ival %% 4
  ival <- ival %/% 4
  ind6 <- ival %% 4
  ival <- ival %/% 4
  ind7 <- ival %% 4

  val0 <- nnames[ind0 + 1]
  val1 <- nnames[ind1 + 1]
  val2 <- nnames[ind2 + 1]
  val3 <- nnames[ind3 + 1]
  val4 <- nnames[ind4 + 1]
  val5 <- nnames[ind5 + 1]
  val6 <- nnames[ind6 + 1]
  val7 <- nnames[ind7 + 1]

  paste(val7, val6, val5, val4, ":", val3, val2, val1, val0, sep="")
}

if (3 == tuple.length) {
  rnames <- apply(as.matrix(c(1:site.cases)), 1, rn3t)
  rownames(tcounts) <- rnames
}
if (4 == tuple.length) {
  rnames <- apply(as.matrix(c(1:site.cases)), 1, rn4t)
  rownames(tcounts) <- rnames
}

cgnam <- function(x) {
  curgap <- sep.lengths[x]
  cn0 <- paste("e.", curgap, sep="")
  cn1 <- paste("s.", curgap, sep="")
  cn2 <- paste("d.", curgap, sep="")
  c(cn0, cn1, cn2)
}
cnames <- apply(as.matrix(c(1:sep.cases)), 1, cgnam)

incprobs <- matrix(0.0, nrow=(dfm.cases + 1), ncol=6)
incprobs[1,] <- 1.0
cvec.uselen <- cvec.maxlen + 1 - sep.min
if (0 < cvec.uselen) {
  prob.bsite <- (0.25) ** (2 * tuple.length)
  prob.zero0 <- (1 - prob.bsite) ** cvec.uselen
  prob.zero1 <- (1 - ((1 + (3 * 2 * tuple.length)) * prob.bsite)) ** cvec.uselen
  prob.zero2 <- (1 - ((1 + (3 * 2 * tuple.length) + (3 * tuple.length * 3 * tuple.length)) * prob.bsite)) ** cvec.uselen

  prob.lzero0 <- log(prob.zero0)
  prob.lzero1 <- log(prob.zero1)
  prob.lzero2 <- log(prob.zero2)
  prob.lone0 <- log(1 - prob.zero0)
  prob.lone1 <- log(1 - prob.zero1)
  prob.lone2 <- log(1 - prob.zero2)

  eprobs0 <- function(x) {
    exp(lchoose(dfm.cases, x) + (x * prob.lone0) + ((dfm.cases - x) * prob.lzero0))
  }
  eprobs1 <- function(x) {
    exp(lchoose(dfm.cases, x) + (x * prob.lone1) + ((dfm.cases - x) * prob.lzero1))
  }
  eprobs2 <- function(x) {
    exp(lchoose(dfm.cases, x) + (x * prob.lone2) + ((dfm.cases - x) * prob.lzero2))
  }
  tmpprobs0 <- apply(as.matrix(1:dfm.cases), 1, eprobs0)
  tmpprobs1 <- apply(as.matrix(1:dfm.cases), 1, eprobs1)
  tmpprobs2 <- apply(as.matrix(1:dfm.cases), 1, eprobs2)

  cprobs0 <- function(x) {
    incprobs[(x + 1), 1] <<- sum(tmpprobs0[x:dfm.cases])
    return()
  }
  cprobs1 <- function(x) {
    incprobs[(x + 1), 2] <<- sum(tmpprobs1[x:dfm.cases])
    return()
  }
  cprobs2 <- function(x) {
    incprobs[(x + 1), 3] <<- sum(tmpprobs2[x:dfm.cases])
    return()
  }
  nt <- apply(as.matrix(1:dfm.cases), 1, cprobs0)
  nt <- apply(as.matrix(1:dfm.cases), 1, cprobs1)
  nt <- apply(as.matrix(1:dfm.cases), 1, cprobs2)


  bprobs <- function(x) {
    curval <- incprobs[x, 1]
    newval <- (1.0 - ((1.0 - curval) ** dfm.cases))
    if (newval < curval) {
      newval <- (curval * dfm.cases)
    }
    if (newval > 1.0) {
      newval <- 1.0
    }
    incprobs[x, 4] <<- newval

    curval <- incprobs[x, 2]
    newval <- (1.0 - ((1.0 - curval) ** dfm.cases))
    if (newval < curval) {
      newval <- (curval * dfm.cases)
    }
    if (newval > 1.0) {
      newval <- 1.0
    }
    incprobs[x, 5] <<- newval

    curval <- incprobs[x, 3]
    newval <- (1.0 - ((1.0 - curval) ** dfm.cases))
    if (newval < curval) {
      newval <- (curval * dfm.cases)
    }
    if (newval > 1.0) {
      newval <- 1.0
    }
    incprobs[x, 6] <<- newval

    return()
  }

  nt <- apply(as.matrix(2:(dfm.cases + 1)), 1, bprobs)

}


options("warn" = -1)
write("#studna blueprint count probability estimates", args[4])

probsline <- c("probs", "e.1", "s.1", "d.1", "e.b", "s.b", "d.b")
write(probsline, args[4], append=TRUE, sep=",", ncolumns=((3 * sep.cases) + 1))

rownames(incprobs) <- c(0:dfm.cases)
write.table(incprobs, args[4], append=TRUE, sep=",", col.names=FALSE, quote=FALSE)

write("#studna blueprint tuplet incidence counts", args[5])

casesline <- cbind("cases", t(as.vector(cnames)))
write(casesline, args[5], append=TRUE, sep=",", ncolumns=((3 * sep.cases) + 1))

write.table(tcounts, args[5], append=TRUE, sep=",", col.names=FALSE, quote=FALSE)



